from copy import deepcopy
import re
import lxml
from typing import Any, List, Tuple, Union
from collections import OrderedDict
from newspaper.configuration import Configuration
import newspaper.parsers as parsers
from newspaper.extractors.defines import AUTHOR_ATTRS, AUTHOR_STOP_WORDS, AUTHOR_VALS


class AuthorsExtractor:
    def __init__(self, config: Configuration) -> None:
        self.config = config
        self.authors: List[str] = []

    def parse(self, doc: lxml.html.Element) -> List[str]:
        """Fetch the authors of the article, return as a list
        Only works for english articles
        """
        _digits = re.compile(r"\d")
        author_stopwords_patt = [re.escape(x) for x in AUTHOR_STOP_WORDS]
        author_stopwords = re.compile(
            r"\b(" + "|".join(author_stopwords_patt) + r")\b", flags=re.IGNORECASE
        )

        def contains_digits(d):
            return bool(_digits.search(d))

        def uniqify_list(lst: List[str]) -> List[str]:
            """Remove duplicates from provided list but maintain original order.
            Ignores trailing spaces and case.

            Args:
                lst (List[str]): Input list of strings, with potential duplicates

            Returns:
                List[str]: Output list of strings, with duplicates removed
            """
            seen = OrderedDict()
            for item in lst:
                seen[item.lower().strip()] = item.strip()
            return [value for item, value in seen.items() if item]

        def parse_byline(search_str):
            """
            Takes a candidate line of html or text and
            extracts out the name(s) in list form:
            >>> parse_byline('<div>By: <strong>Lucas Ou-Yang</strong>,
                    <strong>Alex Smith</strong></div>')
            ['Lucas Ou-Yang', 'Alex Smith']
            """
            # Remove HTML boilerplate
            search_str = re.sub("<[^<]+?>", "", search_str)
            search_str = re.sub("[\n\t\r\xa0]", " ", search_str)

            # Remove original By statement
            m = re.search(r"\b(by|from)[:\s](.*)", search_str, flags=re.IGNORECASE)
            if m:
                search_str = m.group(2)

            search_str = search_str.strip()

            # Chunk the line by non alphanumeric
            # tokens (few name exceptions)
            # >>> re.split("[^\w\'\-\.]",
            #           "Tyler G. Jones, Lucas Ou, Dean O'Brian and Ronald")
            # ['Tyler', 'G.', 'Jones', '', 'Lucas', 'Ou', '',
            #           'Dean', "O'Brian", 'and', 'Ronald']
            name_tokens = re.split(r"[·,\|]|\sand\s|\set\s|\sund\s|/", search_str)
            # some sanity checks
            name_tokens = [s.strip() for s in name_tokens if not contains_digits(s)]
            name_tokens = [s for s in name_tokens if 5 > len(re.findall(r"\w+", s)) > 1]

            return name_tokens

        # Try 1: Search popular author tags for authors

        matches = []
        authors = []

        json_ld_scripts = parsers.get_ld_json_object(doc)

        def get_authors(vals):
            if isinstance(vals, dict):
                if isinstance(vals.get("name"), str):
                    authors.append(vals.get("name"))
                elif isinstance(vals.get("name"), list):
                    authors.extend(vals.get("name"))
            elif isinstance(vals, list):
                for val in vals:
                    if isinstance(val, dict):
                        authors.append(val.get("name"))
                    elif isinstance(val, str):
                        authors.append(val)
            elif isinstance(vals, str):
                authors.append(vals)

        for script_tag in json_ld_scripts:
            if "@graph" in script_tag:
                g = script_tag.get("@graph", [])
                for item in g:
                    if not isinstance(item, dict):
                        continue
                    if item.get("@type") == "Person":
                        authors.append(item.get("name"))
                    if "author" in item:
                        get_authors(item["author"])
            else:
                if "author" in script_tag:
                    get_authors(script_tag["author"])

        def get_text_from_element(node: lxml.html.HtmlElement) -> str:
            """Return the text from an element, including the text from its children
            Args:
                node (lxml.html.HtmlElement): Input node
            Returns:
                str: Text from the node
            """
            if node is None:
                return ""
            if node.tag in ["script", "style", "time"]:
                return ""

            node = deepcopy(node)
            for tag in ["script", "style", "time"]:
                for el in node.xpath(f".//{tag}"):
                    el.getparent().remove(el)
            text = parsers.get_text(node)
            return text

        authors = [re.sub("[\n\t\r\xa0]", " ", x) for x in authors if x]
        doc_root = doc.getroottree()

        def getpath(node):
            if doc_root is not None:
                return doc_root.getpath(node)

        # TODO: be more specific, not a combination of all attributes and values
        for attr in AUTHOR_ATTRS:
            for val in AUTHOR_VALS:
                # found = doc.xpath('//*[@%s="%s"]' % (attr, val))
                found = parsers.get_elements_by_attribs(doc, attribs={attr: val})
                matches.extend([(found, getpath(found)) for found in found])

        matches.sort(
            key=lambda x: x[1], reverse=True
        )  # sort by xpath. we want the most specific match
        matches_reduced: List[Tuple[Any, str]] = []
        for m in matches:
            if len(matches_reduced) == 0:
                matches_reduced.append(m)
            elif not matches_reduced[-1][1].startswith(
                m[1]
            ):  # remove parents of previous node
                matches_reduced.append(m)
        matches_reduced.sort(
            key=lambda x: x[1]
        )  # Preserve some sort of order for the authors

        for match, _ in matches_reduced:
            content: Union[str, List] = ""
            if match.tag == "meta":
                mm = match.xpath("@content")
                if len(mm) > 0:
                    content = mm[0]
            else:
                # TODO: ignore <time> tags, or tags with "on ..."
                # TODO: remove <style> tags https://washingtonindependent.com/how-to-apply-for-reseller-permit-in-washington-state/
                content = get_text_from_element(match)
            if len(content) > 0:
                authors.extend(parse_byline(content))

        # Clean up authors of stopwords such as Reporter, Senior Reporter
        authors = [re.sub(author_stopwords, "", x).strip(" .,-/") for x in authors]
        self.authors = uniqify_list(authors)

        return self.authors



import re
import lxml
import tldextract
from typing import Any, Dict, List, Optional, Tuple
from newspaper import urls
from newspaper.configuration import Configuration
from newspaper.extractors.defines import url_stopwords, category_url_prefixes
import newspaper.parsers as parsers


class CategoryExtractor:
    def __init__(self, config: Configuration) -> None:
        self.config = config
        self.categories: List[str] = []

    def parse(self, source_url: str, doc: lxml.html.Element) -> List[str]:
        """Inputs source lxml root and source url, extracts domain and
        finds all of the top level urls, we are assuming that these are
        the category urls.
        cnn.com --> [cnn.com/latest, world.cnn.com, cnn.com/asia]
        """
        domain_tld = tldextract.extract(source_url)

        links_in_doc = set([a.get("href") for a in parsers.get_tags(doc, tag="a")])

        category_candidates: List[Any] = []

        for p_url in links_in_doc:
            ok, parsed_url = self.is_valid_link(p_url, domain_tld.domain)
            if ok:
                if not parsed_url["domain"]:
                    parsed_url["domain"] = urls.get_domain(
                        source_url, allow_fragments=False
                    )
                    parsed_url["scheme"] = urls.get_scheme(
                        source_url, allow_fragments=False
                    )
                    parsed_url["tld"] = domain_tld

                category_candidates.append(parsed_url)

        _valid_categories = []

        # TODO Stop spamming urlparse and tldextract calls...
        stop_words = set(url_stopwords)
        for p_url in category_candidates:
            path = p_url["path"].lower().split("/")
            subdomain = p_url["tld"].subdomain.lower().split(".")

            conjunction = set(path + subdomain)
            if len(conjunction.intersection(stop_words)) == 0:
                p_url["scheme"] = p_url["scheme"] if p_url["scheme"] else "http"
                if p_url["path"].endswith("/"):
                    p_url["path"] = p_url["path"][:-1]
                _valid_categories.append(
                    p_url["scheme"] + "://" + p_url["domain"] + p_url["path"]
                )

        if len(_valid_categories) == 0:
            other_links_in_doc = self._get_other_links(
                doc, filter_tld=domain_tld.domain
            )
            other_links_in_doc = set(other_links_in_doc)
            for p_url in other_links_in_doc:
                ok, parsed_url = self.is_valid_link(p_url, domain_tld.domain)
                if ok:
                    path = parsed_url["path"].lower().split("/")
                    subdomain = parsed_url["tld"].subdomain.lower().split(".")
                    conjunction = set(path + subdomain)

                    if len(conjunction.intersection(stop_words)) == 0:
                        _valid_categories.append(
                            parsed_url["scheme"]
                            + "://"
                            + parsed_url["domain"]
                            + parsed_url["path"]
                        )

        _valid_categories.append("/")  # add the root

        _valid_categories = list(set(_valid_categories))

        category_urls = [
            urls.prepare_url(p_url, source_url)
            for p_url in _valid_categories
            if p_url is not None
        ]

        self.categories = sorted(category_urls)
        return self.categories

    def _get_other_links(
        self, doc: lxml.html.Element, filter_tld: Optional[str] = None
    ) -> List[str]:
        """Return all links that are not as <a> tags. These can be
        links in javascript tags, json objects, etc.
        """
        html = parsers.node_to_string(doc)
        candidates = re.findall(r'"(https?:\\?/\\?/[^"]*)"', html)

        candidates = [c.replace(r"\/", "/") for c in candidates]
        candidates = [c.replace(r"/\\", "/") for c in candidates]

        def _filter(candidate):
            if filter_tld is not None:
                candidate_tld = tldextract.extract(candidate)
                if candidate_tld.domain != filter_tld:
                    return False
            if re.search(r"\.(css|js|json|xml|rss|jpg|jpeg|png|)$", candidate, re.I):
                return False

            path = urls.get_path(candidate, allow_fragments=False)
            path_chunks = [x for x in path.split("/") if len(x) > 0]
            if "index.html" in path_chunks:
                path_chunks.remove("index.html")

            if len(path_chunks) > 2 or len(path_chunks) == 0:
                return False

            return True

        return filter(_filter, candidates)

    def is_valid_link(self, url: str, filter_tld: str) -> Tuple[bool, Dict[str, Any]]:
        """Is the url a possible category?"""
        parsed_url = {
            "scheme": urls.get_scheme(url, allow_fragments=False),
            "domain": urls.get_domain(url, allow_fragments=False),
            "path": urls.get_path(url, allow_fragments=False),
            "tld": None,
        }

        # No domain or path
        if not parsed_url["domain"] and not parsed_url["path"]:
            return False, parsed_url
        # remove any url that starts with #
        if parsed_url["path"] and parsed_url["path"].startswith("#"):
            return False, parsed_url
        # remove urls that are not http or https (ex. mailto:)
        if parsed_url["scheme"] and (
            parsed_url["scheme"] != "http" and parsed_url["scheme"] != "https"
        ):
            return False, parsed_url

        path_chunks = [x for x in parsed_url["path"].split("/") if len(x) > 0]
        if "index.html" in path_chunks:
            path_chunks.remove("index.html")

        if parsed_url["domain"]:
            child_tld = tldextract.extract(url)
            parsed_url["tld"] = child_tld
            child_subdomain_parts = child_tld.subdomain.split(".")

            # Ex. microsoft.com is definitely not related to
            # espn.com, but espn.go.com is probably related to espn.com
            if (
                child_tld.domain != filter_tld
                and filter_tld not in child_subdomain_parts
            ):
                return False, parsed_url

            if child_tld.subdomain in ["m", "i"]:
                return False, parsed_url

            subd = "" if child_tld.subdomain == "www" else child_tld.subdomain

            if len(subd) > 0 and len(path_chunks) == 0:
                return True, parsed_url  # Allow http://category.domain.tld/

        # we want a path with just one subdir
        # cnn.com/world and cnn.com/world/ are both valid_categories
        # cnn.com/world/europe is not a valid_category
        # europe.cnn.com/economy is a valid_category
        if len(path_chunks) > 2 or len(path_chunks) == 0:
            return False, parsed_url

        if any(
            [x.startswith("_") or x.startswith("#") for x in path_chunks]
        ):  # Ex. cnn.com/_static/
            return False, parsed_url

        if len(path_chunks) == 2 and path_chunks[0] in category_url_prefixes:
            return True, parsed_url

        return len(path_chunks) == 1 and 1 < len(path_chunks[0]) < 20, parsed_url



import logging
import urllib.parse
from copy import copy
import re
from typing import List, Optional, Tuple
import lxml
from PIL import Image, ImageFile
import requests
from newspaper import urls
import newspaper.parsers as parsers
from newspaper.configuration import Configuration
import newspaper.extractors.defines as defines
from newspaper.urls import urljoin_if_valid

log = logging.getLogger(__name__)


class ImageExtractor:
    """Extractor class for images in articles. Getting top image,
    image list, favicon, etc."""

    def __init__(self, config: Configuration) -> None:
        self.config = config
        self.top_image: Optional[str] = None
        self.meta_image: Optional[str] = None
        self.images: List[str] = []
        self.favicon: Optional[str] = None
        self._chunksize = 1024

    def parse(
        self, doc: lxml.html.Element, top_node: lxml.html.Element, article_url: str
    ) -> None:
        """main method to extract images from a document

        Args:
            doc (lxml.html.Element): _description_
        """
        self.favicon = self._get_favicon(doc)

        self.meta_image = self._get_meta_image(doc)
        if self.meta_image:
            self.meta_image = urljoin_if_valid(article_url, self.meta_image)
        self.images = [
            urljoin_if_valid(article_url, u)
            for u in self._get_images(doc)  # Tried to use top_node, but images
            # were not found in some cases (times_001.html)
            if u and u.strip()
        ]
        self.top_image = self._get_top_image(doc, top_node, article_url)

    def _get_favicon(self, doc: lxml.html.Element) -> str:
        """Extract the favicon from a website http://en.wikipedia.org/wiki/Favicon
        <link rel="shortcut icon" type="image/png" href="favicon.png" />
        <link rel="icon" type="image/png" href="favicon.png" />
        """
        meta = parsers.get_tags(
            doc, tag="link", attribs={"rel": "icon"}, attribs_match="substring"
        )
        if meta:
            favicon = parsers.get_attribute(meta[0], "href")
            return favicon
        return ""

    def _get_meta_image(self, doc: lxml.html.Element) -> str:
        """Extract image from the meta tags of the document."""
        candidates: List[Tuple[str, int]] = []
        for elem in defines.META_IMAGE_TAGS:
            if "|" in elem["value"]:
                items = parsers.get_tags_regex(
                    doc, tag=elem["tag"], attribs={elem["attr"]: elem["value"]}
                )
            else:
                items = parsers.get_tags(
                    doc,
                    tag=elem["tag"],
                    attribs={elem["attr"]: elem["value"]},
                    attribs_match="exact",
                )

            candidates.extend((el.get(elem["content"]), elem["score"]) for el in items)

        candidates = [c for c in candidates if c[0]]

        candidates.sort(key=lambda x: x[1], reverse=True)

        return candidates[0][0] if candidates else ""

    def _get_images(self, doc: lxml.html.Element) -> List[str]:
        def get_src(image):
            # account for src, data-src and other attributes
            srcs = [image.attrib.get(x) for x in image.attrib if "src" in x]
            srcs = [x for x in srcs if x and not x.startswith("data:")]

            srcs.sort(key=lambda x: 0 if x.lower().startswith("http") else 1)

            return srcs[0] if srcs else None

        images = [get_src(x) for x in parsers.get_tags(doc, tag="img")]
        images = [x for x in images if x]

        return images

    def _get_top_image(
        self, doc: lxml.html.Element, top_node: lxml.html.Element, article_url: str
    ) -> str:
        def node_distance(node1, node2):
            path1 = node1.getroottree().getpath(node1).split("/")
            path2 = node2.getroottree().getpath(node2).split("/")
            for i, (step1, step2) in enumerate(zip(path1, path2)):
                if step1 != step2:
                    return len(path1[i:]) + len(path2[i:])

            return abs(len(path1) - len(path2))

        if self.meta_image:
            if not self.config.fetch_images or self._check_image_size(
                self.meta_image, article_url
            ):
                return self.meta_image

        img_cand = []
        for img in parsers.get_tags(doc, tag="img"):
            if not img.get("src"):
                continue
            if img.get("src").startswith("data:"):
                continue

            if top_node is not None:
                distance = node_distance(top_node, img)
                img_cand.append((img, distance))
            else:
                if self._check_image_size(img.get("src"), article_url):
                    return img.get("src")

        img_cand.sort(key=lambda x: x[1])

        for img in img_cand:
            if self._check_image_size(img[0].get("src"), article_url):
                return img[0].get("src")

        return ""

    def _check_image_size(self, url: str, referer: Optional[str]) -> bool:
        img = self._fetch_image(
            url,
            referer,
        )
        if not img:
            return False

        width, height = img.size

        if self.config.top_image_settings["min_width"] > width:
            return False
        if self.config.top_image_settings["min_height"] > height:
            return False
        if self.config.top_image_settings["min_area"] > width * height:
            return False

        if (
            re.search(r"(logo|sprite)", url, re.IGNORECASE)
            and self.config.top_image_settings["min_area"] > width * height / 10
        ):
            return False

        return True

    def _fetch_image(self, url: str, referer: Optional[str]) -> Optional[Image.Image]:
        def clean_url(url):
            """Url quotes unicode data out of urls"""
            if not isinstance(url, str):
                return url

            url = url.encode("utf8")
            url = "".join(
                [
                    urllib.parse.quote(c) if ord(c) >= 127 else c
                    for c in url.decode("utf-8")
                ]
            )
            return url

        requests_params = copy(self.config.requests_params)
        requests_params["headers"]["Referer"] = referer
        max_retries = self.config.top_image_settings["max_retries"]

        cur_try = 0
        url = clean_url(url)
        if not url or not url.startswith(("http://", "https://")):
            return None

        response = None
        while True:
            try:
                response = requests.get(
                    url,
                    stream=True,
                    **requests_params,
                )

                content_type = response.headers.get("Content-Type")

                if not content_type or "image" not in content_type.lower():
                    return None

                p = ImageFile.Parser()
                new_data = response.raw.read(self._chunksize)
                while not p.image and new_data:
                    try:
                        p.feed(new_data)
                    except (IOError, ValueError) as e:
                        log.warning(
                            "error %s while fetching: %s refer: %s",
                            str(e),
                            url,
                            requests_params["headers"].get("Referer"),
                        )
                        return None
                    except Exception as e:
                        # For some favicon.ico images, the image is so small
                        # that our PIL feed() method fails a length test.
                        is_favicon = urls.url_to_filetype(url) == "ico"
                        if not is_favicon:
                            raise e
                        return None
                    new_data = response.raw.read(self._chunksize)
                return p.image
            except requests.exceptions.RequestException:
                cur_try += 1
                if cur_try >= max_retries:
                    log.warning(
                        "error while fetching: %s refer: %s",
                        url,
                        requests_params["headers"].get("Referer"),
                    )
                    return None
            finally:
                if response is not None:
                    response.raw.close()
                    if response.raw._connection:
                        response.raw._connection.close()



# -*- coding: utf-8 -*-
# Much of the code here was forked from https://github.com/codelucas/newspaper
# Copyright (c) Lucas Ou-Yang (codelucas)

from dataclasses import dataclass
from pathlib import Path
import pickle
from typing import Callable, Optional
import hashlib
import time

from newspaper.settings import CACHE_DIRECTORY


@dataclass
class Video:
    """Video object"""

    embed_type: Optional[str] = None
    # Video provider name
    provider: Optional[str] = None
    width: Optional[int] = None
    height: Optional[int] = None
    # embedding html code
    embed_code: Optional[str] = None
    src: Optional[str] = None


class ParsingCandidate:
    def __init__(self, url, link_hash):
        self.url = url
        self.link_hash = link_hash


class RawHelper:
    @staticmethod
    def get_parsing_candidate(url, raw_html):
        if isinstance(raw_html, str):
            raw_html = raw_html.encode("utf-8", "replace")
        link_hash = "%s.%s" % (hashlib.md5(raw_html).hexdigest(), time.time())
        return ParsingCandidate(url, link_hash)


class URLHelper:
    @staticmethod
    def get_parsing_candidate(url_to_crawl):
        # Replace shebang in urls
        final_url = (
            url_to_crawl.replace("#!", "?_escaped_fragment_=")
            if "#!" in url_to_crawl
            else url_to_crawl
        )
        link_hash = "%s.%s" % (hashlib.md5(final_url).hexdigest(), time.time())
        return ParsingCandidate(final_url, link_hash)


class CacheDiskDecorator:
    def __init__(self, enabled=True):
        self._enabled = enabled
        self._seconds = 86400
        self._cache_folder = CACHE_DIRECTORY

    @property
    def enabled(self):
        return self._enabled

    @enabled.setter
    def enabled(self, value):
        self._enabled = value

    def get_cache_file(self, domain):
        filename = hashlib.sha1((domain).encode("utf-8")).hexdigest()
        return Path(self._cache_folder) / filename

    def _do_cache(self, target_function) -> Callable:
        def inner_function(*args, **kwargs):
            """Calculate a cache key based on the decorated method signature
            args[1] indicates the domain of the inputs, we hash on domain!
            """
            if not self.enabled:
                return target_function(*args, **kwargs)

            filepath = self.get_cache_file(kwargs.get("domain") or args[1])

            # verify that the cached object exists and is less than
            # X seconds old
            if filepath.exists():
                modified = filepath.stat().st_mtime
                age_seconds = time.time() - modified
                if age_seconds < self._seconds:
                    with open(filepath, "rb") as f:
                        return pickle.load(f)

            # call the decorated function...
            result = target_function(*args, **kwargs)
            # ... and save the cached object for next time
            with open(filepath, "wb") as f:
                pickle.dump(result, f)

            return result

        return inner_function

    def __call__(self, seconds=None):
        self._seconds = seconds or self._seconds
        return self._do_cache



# -*- coding: utf-8 -*-
# Much of the code here was forked from https://github.com/codelucas/newspaper
# Copyright (c) Lucas Ou-Yang (codelucas)

"""
Holds misc. utility methods which prove to be
useful throughout this library.
"""

import logging
from pathlib import Path
import random
import string
import sys
import time

from bs4 import BeautifulSoup

from newspaper.languages import get_language_from_iso639_1
from newspaper import settings
from .classes import CacheDiskDecorator, URLHelper, RawHelper

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)

cache_disk = CacheDiskDecorator(enabled=True)


def domain_to_filename(domain):
    """All '/' are turned into '-', no trailing. schema's
    are gone, only the raw domain + ".txt" remains
    """
    filename = domain.replace("/", "-")
    if filename[-1] == "-":
        filename = filename[:-1]
    filename += ".txt"
    return filename


def filename_to_domain(filename):
    """[:-4] for the .txt at end"""
    return filename.replace("-", "/")[:-4]


def extract_meta_refresh(html):
    """Parses html for a tag like:
    <meta http-equiv="refresh" content="0;
            URL='http://sfbay.craigslist.org/eby/cto/5617800926.html'" />
    Example can be found at:
        https://www.google.com/url?rct=j&sa=t&url=
        http://sfbay.craigslist.org/eby/cto/5617800926.html&
        ct=ga&cd=CAAYATIaYTc4ZTgzYjAwOTAwY2M4Yjpjb206ZW46VVM&
        usg=AFQjCNF7zAl6JPuEsV4PbEzBomJTUpX4Lg
    """
    soup = BeautifulSoup(html, "html.parser")
    element = soup.find("meta", attrs={"http-equiv": "refresh"})
    if element:
        try:
            _, url_part = element["content"].split(";")
        except ValueError:
            # In case there are not enough values to unpack
            # for instance: <meta http-equiv="refresh" content="600" />
            return None
        else:
            # Get rid of any " or ' inside the element
            # for instance:
            # <meta http-equiv="refresh" content="0;
            #           URL='http://sfbay.craigslist.org/eby/cto/5617800926.html'" />
            if url_part.lower().startswith("url="):
                return url_part[4:].replace('"', "").replace("'", "")


def to_valid_filename(s):
    """Converts arbitrary string (for us domain name)
    into a valid file name for caching
    """
    valid_chars = "-_.() %s%s" % (string.ascii_letters, string.digits)
    return "".join(c for c in s if c in valid_chars)


def clear_memo_cache(source):
    """Clears the memoization cache for this specific news domain"""
    cache_file = settings.MEMO_DIR / domain_to_filename(source.domain)
    if cache_file.exists():
        cache_file.unlink()
    else:
        log.info("memo file for %s has already been deleted!", source.domain)


def memorize_articles(source, articles):
    """When we parse the <a> links in an <html> page, on the 2nd run
    and later, check the <a> links of previous runs. If they match,
    it means the link must not be an article, because article urls
    change as time passes. This method also uniquifies articles.
    """
    if len(articles) == 0:
        return []

    source_domain = source.domain

    cache_file = settings.MEMO_DIR / domain_to_filename(source_domain)

    if cache_file.exists():
        with open(cache_file, "r", encoding="utf-8") as f:
            urls = f.readlines()

        valid_urls = [u.strip() for u in urls if u.strip()]

        # select not already seen urls
        cur_articles = {
            article.url: article
            for article in articles
            if article.url not in valid_urls
        }

        valid_urls.extend([url for url in cur_articles])

    else:
        cur_articles = {article.url: article for article in articles}
        valid_urls = list(cur_articles.keys())

    if len(valid_urls) > source.config.max_file_memo:
        valid_urls = valid_urls[: source.config.max_file_memo]
        log.warning("Source %s: memorization file overflow, truncating", source.domain)

    with open(cache_file, "w", encoding="utf-8") as f:
        f.writelines([x + "\n" for x in valid_urls if x])

    return list(cur_articles.values())


def get_useragent():
    """Returns a random useragent from our saved file"""
    with open(settings.USERAGENTS, "r", encoding="utf-8") as f:
        agents = f.readlines()
        selection = random.randint(0, len(agents) - 1)
        agent = agents[selection]
        return agent.strip()


def get_available_languages():
    """Returns a list of available languages and their 2 char input codes"""
    stopword_files = Path(settings.STOPWORDS_DIR).glob("stopwords-??.txt")
    for file in stopword_files:
        yield file.stem.split("-")[1]


def print_available_languages():
    """Prints available languages with their full names"""
    print("\nYour available languages are:")
    print("\ninput code\t\tfull name")

    for lang in get_available_languages():
        print("  %s\t\t\t  %s" % (lang, get_language_from_iso639_1(lang)))

    print()


def progressbar(it, prefix="", size=60, out=sys.stdout):
    """Display a simple progress bar without
    heavy dependencies like tqdm"""
    count = len(it)
    start = time.time()

    def show(j):
        x = int(size * j / count)
        remaining = ((time.time() - start) / j) * (count - j)

        mins, sec = divmod(remaining, 60)
        time_str = f"{int(mins):02}:{sec:05.2f}"

        print(
            f"{prefix}[{u'█'*x}{('.'*(size-x))}] {j}/{count} Est wait {time_str}",
            end="\r",
            file=out,
            flush=True,
        )

    for i, item in enumerate(it):
        yield item
        show(i + 1)
    print("\n", flush=True, file=out)


def print_node_tree(node, header="", last=True, with_gravity=True):
    """Prints out the html node tree for nodes with gravity scores
    debugging method
    """
    elbow = "└──"
    pipe = "│  "
    tee = "├──"
    if not with_gravity or node.get("gravityScore"):
        node_attribs = {
            k: node.attrib.get(k) for k in ["class", "id"] if node.attrib.get(k)
        }
        score = float(node.get("gravityScore", 0))
        print(
            header
            + (elbow if last else tee)
            + node.tag
            + f"({score:0.1f}) {node_attribs}"
        )
        blank = "   "
    else:
        blank = ""

    children = list(node.iterchildren())
    for i, c in enumerate(children):
        print_node_tree(
            c, header=header + (blank if last else pipe), last=i == len(children) - 1
        )


__all__ = [
    "RawHelper",
    "URLHelper",
    "domain_to_filename",
    "filename_to_domain",
    "extract_meta_refresh",
    "to_valid_filename",
    "cache_disk",
    "clear_memo_cache",
    "memorize_articles",
    "get_useragent",
    "get_available_languages",
    "print_available_languages",
    "progressbar",
    "print_node_tree",
]



# -*- coding: utf-8 -*-
# Much of the code here was forked from https://github.com/codelucas/newspaper
# Copyright (c) Lucas Ou-Yang (codelucas)

"""
Module provinding the OutputFormatter class, which converts the article top node
to plain text, removing most boilerplate and other unwanted elements.
"""
from copy import deepcopy
import logging
import re
from statistics import mean, stdev
from typing import Any, Dict, Tuple

import lxml
from newspaper import parsers
from newspaper.configuration import Configuration
from newspaper import settings

log = logging.getLogger(__name__)

WHITESPACE_CHARS = "\n\r\t " + "\u00a0" + "\ufeff"
MAX_PARAGRAPH_BEFORE_TITLE = 200


class OutputFormatter:
    """Class that converts the article top node into text, cleaning up
    debris tags, replacing <br> with newlines, etc.

    if `config.clean_article_html` is True, then the article's html is
    cleaned as well. Only `settings.CLEAN_ARTICLE_TAGS` are allowed to
    remain in the html.
    """

    def __init__(self, config=None):
        self.config = config or Configuration()

    def get_formatted(
        self, top_node: lxml.html.HtmlElement, article_title: str = None
    ) -> Tuple[str, str]:
        """Returns the body text of an article, and also the cleaned html body
        article of the article.
        Arguments:
            top_node {lxml.html.HtmlElement} -- The top node element of the article
            article_title {str} -- The title of the article, if available, to
                be removed from the text (and max 1 paragraph before it)
        Returns:
            Tuple[str, str] -- The body text of the article, and the cleaned
            html body of the article
        """
        html, text = "", ""
        if top_node is None:
            return (text, html)

        node_cleaned = deepcopy(top_node)

        self._remove_negativescores_nodes(node_cleaned)

        if not self.config.clean_article_html:
            # We deliver the HTML untouched (only the negative nodes are removed)
            html = parsers.node_to_string(node_cleaned)

        self._remove_advertisement_nodes(node_cleaned)

        self._remove_unlikely_nodes(node_cleaned)

        self._remove_empty_tags(node_cleaned)

        # removes some same level tags that might
        # contain non-content like menus, gallery,  etc.
        # this can misfire on some sites
        self._remove_trailing_media_div(node_cleaned)

        if self.config.clean_article_html:
            html = self._create_clean_html(node_cleaned)

        text = self._convert_to_text(node_cleaned, article_title)

        return (text, html)

    def _convert_to_text(
        self, top_node: lxml.html.HtmlElement, article_title: str = None
    ) -> str:
        article_cleaner = lxml.html.clean.Cleaner()
        article_cleaner.javascript = True
        article_cleaner.style = True
        article_cleaner.remove_unknown_tags = False
        article_cleaner.meta = True
        article_cleaner.embedded = True
        article_cleaner.frames = True
        article_cleaner.allow_tags = settings.BLOCK_LEVEL_TAGS + ["br"]

        cleaned_node = article_cleaner.clean_html(top_node)
        # TODO: do not remove newlines in <pre> tags

        txts = [
            re.sub(r"[\s\t\xa0\uFEFF]+", " ", value, flags=re.UNICODE)
            for value in cleaned_node.itertext()
        ]
        txts = [x.strip(" \t") for x in txts if x.strip(WHITESPACE_CHARS)]
        if article_title and len(txts) > 1:
            # Remove the title and the first paragraph before it
            # (if it's not too long)
            def normalize_string(s: str) -> str:
                # remove punctuation, double spaces and lowers the case
                s = re.sub(r"[^\w\s]", "", s)
                s = re.sub(r"\s+", " ", s)
                s = s.lower()
                return s

            if normalize_string(txts[0]) == normalize_string(article_title):
                txts = txts[1:]
            elif len(txts[0]) < MAX_PARAGRAPH_BEFORE_TITLE and normalize_string(
                txts[1]
            ) == normalize_string(article_title):
                txts = txts[2:]

        return "\n\n".join(txts)

    def _create_clean_html(self, top_node: lxml.html.HtmlElement):
        article_cleaner = lxml.html.clean.Cleaner()
        article_cleaner.javascript = True
        article_cleaner.style = True
        article_cleaner.remove_unknown_tags = False
        article_cleaner.meta = True
        article_cleaner.embedded = True

        article_cleaner.allow_tags = settings.CLEAN_ARTICLE_TAGS

        cleaned_node = article_cleaner.clean_html(top_node)
        return parsers.node_to_string(cleaned_node)

    def _add_newline_to_br(self, top_node: lxml.html.HtmlElement):
        """Replace all br tags in 'element' with a newline character"""
        br_tags = top_node.xpath(".//br")
        for br in br_tags:
            br.tail = "\n" + br.tail if br.tail else "\n"

    def _remove_negativescores_nodes(self, top_node: lxml.html.HtmlElement):
        """If there are elements inside our top node that have a
        negative gravity score, let's give em the boot.
        """
        gravity_items = top_node.xpath(".//*[@gravityScore]")
        for item in gravity_items:
            score = item.attrib.get("gravityScore", "0")
            score = float(score)
            if score < 1:
                item.getparent().remove(item)

    def _remove_empty_tags(self, top_node: lxml.html.HtmlElement):
        """It's common in top_node to have tags that are filled with data
        in their properties but do not have any displayable text.
        """
        all_nodes = parsers.get_tags(top_node)
        all_nodes.reverse()
        for el in all_nodes:
            tag = el.tag
            if tag == "br":
                continue
            if len(parsers.get_elements_by_tagslist(el, ["object", "embed"])) > 0:
                continue

            txt = parsers.get_text(el)
            txt = re.sub(r"[\s\t]+", "", txt)

            if not txt:
                parsers.remove(el)

    def _get_top_level_nodes(self, top_node: lxml.html.HtmlElement):
        """Returns a list of nodes that are of the top level"""
        top_level_nodes = top_node.getchildren()
        if top_node.tag == "body" and len(top_level_nodes) == 1:
            top_level_nodes = top_level_nodes[0].getchildren()

        return top_level_nodes

    def _remove_trailing_media_div(self, top_node: lxml.html.HtmlElement):
        """Punish the *last top level* node in the top_node if it's
        DOM depth is too deep or has a a lot of links. Many media non-content
        links are eliminated: "related", "loading gallery", etc. It skips
        removal if last top level node's class is one of NON_MEDIA_CLASSES.
        """

        NON_MEDIA_CLASSES = ("zn-body__read-all",)

        top_level_nodes = self._get_top_level_nodes(top_node)

        if len(top_level_nodes) < 3:
            return

        last_node = top_level_nodes[-1]

        last_node_class = parsers.get_attribute(last_node, "class")
        if last_node_class in NON_MEDIA_CLASSES:
            return
        if last_node.tag != "p" and len(parsers.get_tags(last_node, "p")) > 0:
            if parsers.get_node_gravity_score(last_node) > 15:
                return

        if parsers.get_node_depth(last_node) >= 2:
            parsers.remove(last_node)
        elif parsers.is_highlink_density(last_node):
            parsers.remove(last_node)

    def _top_nodes_stats(self, top_node: lxml.html.HtmlElement):
        """Returns a list of top nodes and stats about them"""
        top_nodes = self._get_top_level_nodes(top_node)
        node_stats: Dict[str, Dict[str, Any]] = {}
        for el in top_nodes:
            node_stats[el.tag] = node_stats.setdefault(
                el.tag, {"count": 0, "gravity": [], "depth": []}
            )
            node_stats[el.tag]["count"] += 1
            node_stats[el.tag]["gravity"].append(parsers.get_node_gravity_score(el))
            node_stats[el.tag]["depth"].append(parsers.get_node_depth(el))

        node_stats = {
            k: {
                "count": v["count"],
                "gravity_mean": mean(v["gravity"]),
                "gravity_std": stdev(v["gravity"]) if len(v["gravity"]) > 1 else 0,
                "depth": mean(v["depth"]),
                "depth_std": stdev(v["depth"]) if len(v["depth"]) > 1 else 0,
            }
            for k, v in node_stats.items()
        }

        return node_stats

    def _remove_unlikely_nodes(self, top_node: lxml.html.HtmlElement):
        """Remove unlikely top level nodes from the top node
        based on statistical analysis based on depth and gravity score
        """
        stats = self._top_nodes_stats(top_node)
        top_nodes = self._get_top_level_nodes(top_node)

        # has p and divs. Analyse if divs are not boilerplate or ads
        if "p" in stats and "div" in stats:
            for node in top_nodes:
                if node.tag != "div":
                    continue
                gravity = parsers.get_node_gravity_score(node)
                depth = parsers.get_node_depth(node)

                if (
                    depth > round(stats["div"]["depth"] + stats["div"]["depth_std"])
                    or depth > round(stats["p"]["depth"] + stats["p"]["depth_std"])
                    or gravity
                    < stats["p"]["gravity_mean"] - 2 * stats["p"]["gravity_std"]
                    or gravity
                    < stats["div"]["gravity_mean"] - 2 * stats["div"]["gravity_std"]
                ):
                    parsers.remove(node)

    def _remove_advertisement_nodes(self, top_node: lxml.html.HtmlElement):
        """Remove nodes that may contain advertisement content."""

        divs = top_node.xpath(".//div")
        stats = self._top_nodes_stats(top_node)

        for el in divs:
            # Does it contain p tags?
            if len(parsers.get_tags(el, "p")):
                if parsers.is_highlink_density(el):
                    gravity = parsers.get_node_gravity_score(el)
                    if len(stats):
                        limit = max(
                            [
                                stats[x]["gravity_mean"] - 2 * stats[x]["gravity_std"]
                                for x in stats
                            ]
                        )
                    else:
                        limit = 15  # no gravity scores, then remove all

                    if gravity < limit:
                        parsers.remove(el)

                continue

            if parsers.is_highlink_density(el):
                parsers.remove(el)
                continue
            attrs = el.get("class", "") + " " + el.get("id", "")
            if re.search(settings.ADVERTISEMENT_ATTR_VALUES, attrs, re.IGNORECASE):
                parsers.remove(el)
                continue



from typing import Dict, List, Union

MOTLEY_REPLACEMENT = ("&#65533;", "")
TITLE_REPLACEMENTS = ("&raquo;", "»")

A_REL_TAG_SELECTOR = "//a[@rel='tag']"
A_HREF_TAG_SELECTOR = (
    "//a[contains(@href, '/tag/')] | //a[contains(@href, '/tags/')] |"
    " //a[contains(@href, '/topic/')] | //a[contains(@href, '?keyword=')]"
)
RE_LANG = r"^[A-Za-z]{2}$"

AUTHOR_ATTRS = ["name", "rel", "itemprop", "class", "id", "property"]
AUTHOR_VALS = [
    "author",
    "byline",
    "dc.creator",
    "byl",
    "article:author",
    "article:author_name",
    "story-byline",
    "article-author",
    "parsely-author",
    "sailthru.author",
    "citation_author",
]

AUTHOR_STOP_WORDS = [
    "By",
    "Reuters",
    "IANS",
    "AP",
    "AFP",
    "PTI",
    "IANS",
    "ANI",
    "DPA",
    "Senior Reporter",
    "Reporter",
    "Writer",
    "Opinion Writer",
]
TITLE_META_INFO = [
    "dc.title",
    "og:title",
    "headline",
    "articletitle",
    "article-title",
    "parsely-title",
    "title",
]
PUBLISH_DATE_META_INFO = [
    "published_date",
    "published_time",
    "cXenseParse:publishtime",
    "pubdate",
    "publish_date",
    "PublishDate",
    "dcterms.created",
    "rnews:datePublished",
    "article:published_time",
    "prism.publicationDate",
    "displaydate",
    "OriginalPublicationDate",
    "og:published_time",
    "datePublished",
    "article_date_original",
    "article.published",
    "published_time_telegram",
    "sailthru.date",
    "date",
    "Date",
    "original-publish-date",
    "DC.date.issued",
    "dc.date",
    "DC.Date",
    "parsely-pub-date",
    "publishtime",
    "og:regDate",
    "publication_date",
    "uploadDate",
    "coverageEndTime",
    "publishdate",
    "publish-date",
    "publishedAtDate",
    "dcterms.date",
    "publishedDate",
    "creationDateTime",
    "pub_date",
    "updated_time",
    "og:updated_time",
    "datemodified",
    "last-modified",
    "Last-Modified",
    "DC.date.modified",
    "article:modified_time",
    "modified_time",
    "modifiedDateTime",
    "dc.dcterms.modified",
    "lastmod",
]

PUBLISH_DATE_TAGS: List[Dict[str, str]] = [
    {"attribute": "itemprop", "value": "datePublished", "content": "datetime"},
    {"attribute": "pubdate", "value": "pubdate", "content": "datetime"},
    {"attribute": "class", "value": "entry-date", "content": "datetime"},
]
ARTICLE_BODY_TAGS: List[Dict[str, Union[str, int]]] = [
    {"tag": "article", "role": "article", "score_boost": 25},
    {"itemprop": "articleBody", "score_boost": 100},
    {"itemprop": "articleText", "score_boost": 40},
    {"itemtype": "https://schema.org/Article", "score_boost": 30},
    {"itemtype": "https://schema.org/NewsArticle", "score_boost": 30},
    {"itemtype": "https://schema.org/BlogPosting", "score_boost": 20},
    {"itemtype": "https://schema.org/ScholarlyArticle", "score_boost": 20},
    {"itemtype": "https://schema.org/SocialMediaPosting", "score_boost": 20},
    {"itemtype": "https://schema.org/TechArticle", "score_boost": 20},
    {
        "class": "re:paragraph|entry-content|article-text|article-body",
        "score_boost": 15,
    },
]
META_IMAGE_TAGS: List[Dict[str, Union[str, int]]] = [
    {
        "tag": "meta",
        "attr": "property",
        "value": "og:image",
        "content": "content",
        "score": 10,
    },
    {
        "tag": "link",
        "attr": "rel",
        "value": "image_src|img_src",
        "content": "href",
        "score": 8,
    },
    {
        "tag": "meta",
        "attr": "name",
        "value": "og:image",
        "content": "content",
        "score": 8,
    },
    {"tag": "link", "attr": "rel", "value": "icon", "content": "href", "score": 5},
]
META_LANGUAGE_TAGS: List[Dict[str, str]] = [
    {"tag": "meta", "attr": "property", "value": "og:locale"},
    {"tag": "meta", "attr": "http-equiv", "value": "content-language"},
    {"tag": "meta", "attr": "name", "value": "lang"},
]

url_stopwords: List[str] = [
    "about",
    "academy",
    "account",
    "admin",
    "advert",
    "advertise",
    "adverts",
    "archive",
    "archives",
    "author",
    "authors",
    "bebo",
    "board",
    "boards",
    "browse",
    "career",
    "careers",
    "charts",
    "comment",
    "comments",
    "contact",
    "contacts",
    "coupons",
    "developer",
    "developers",
    "donate",
    "download",
    "downloads",
    "edit",
    "event",
    "events",
    "facebook",
    "faq",
    "faqs",
    "feed",
    "feeds",
    "feedback",
    "flickr",
    "forum",
    "forums",
    "friendster",
    "help",
    "how-to",
    "howto",
    "imgur",
    "info",
    "itunes",
    "jobs",
    "legal",
    "linkedin",
    "login",
    "mail",
    "map",
    "maps",
    "mobile",
    "myspace",
    "newsletter",
    "newsletters",
    "password",
    "plus",
    "preference",
    "preferences",
    "privacy",
    "privacy-policy",
    "product",
    "products",
    "profile",
    "profiles",
    "proxy",
    "purchase",
    "register",
    "search",
    "services",
    "shop",
    "shopping",
    "signup",
    "site-map",
    "siteindex",
    "sitemap",
    "sitemaps",
    "stop",
    "store",
    "stumbleupon",
    "subscribe",
    "subscription",
    "tag",
    "terms",
    "tickets",
    "toc",
    "twitter",
    "upload",
    "uploads",
    "vimeo",
    "wp-admin",
    "wp-includes",
    "wp-content",
    "wp-json",
    "wp-login",
    "xmlrpc",
    "xmlrpc.php",
    "youtube",
]

category_url_prefixes: List[str] = [
    "category",
    "categories",
    "topic",
    "topics",
    "subject",
    "subjects",
    "section",
    "sections",
    "cat",
    "cats",
]



from datetime import datetime
import re
from typing import Optional

import lxml
from newspaper import urls
from newspaper.configuration import Configuration
import newspaper.parsers as parsers
from dateutil.parser import parse as date_parser

from newspaper.extractors.defines import PUBLISH_DATE_META_INFO, PUBLISH_DATE_TAGS


class PubdateExtractor:
    def __init__(self, config: Configuration) -> None:
        self.config = config
        self.pubdate: Optional[datetime] = None

    def parse(self, article_url: str, doc: lxml.html.Element) -> Optional[datetime]:
        """3 strategies for publishing date extraction. The strategies
        are descending in accuracy and the next strategy is only
        attempted if a preferred one fails.

        1. Pubdate from URL
        2. Pubdate from metadata
        3. Raw regex searches in the HTML + added heuristics
        """

        def parse_date_str(date_str):
            if date_str:
                try:
                    self.pubdate = date_parser(date_str)
                    return self.pubdate
                except (ValueError, OverflowError, AttributeError, TypeError):
                    # near all parse failures are due to URL dates without a day
                    # specifier, e.g. /2014/04/
                    return None

        date_matches = []
        date_match = re.search(urls.STRICT_DATE_REGEX, article_url)
        if date_match:
            date_str = date_match.group(0)
            datetime_obj = parse_date_str(date_str)
            if datetime_obj:
                date_matches.append((datetime_obj, 10))  # date and matchscore

        # yoast seo structured data or json-ld
        json_ld_scripts = parsers.get_ld_json_object(doc)

        for script_tag in json_ld_scripts:
            if "@graph" in script_tag:
                g = script_tag.get("@graph", [])
                for item in g:
                    if not isinstance(item, dict):
                        continue
                    date_str = item.get("datePublished")
                    datetime_obj = parse_date_str(date_str)
                    if datetime_obj:
                        date_matches.append((datetime_obj, 10))
            else:
                for k in script_tag:
                    if k in ["datePublished", "dateCreated"]:
                        date_str = script_tag.get(k)
                        datetime_obj = parse_date_str(date_str)
                        if datetime_obj:
                            date_matches.append((datetime_obj, 9))

        # get <time> tags
        for item in parsers.get_tags(doc, tag="time"):
            if item.get("datetime"):
                date_str = item.get("datetime")
                datetime_obj = parse_date_str(date_str)
                if datetime_obj:
                    if item.text and re.search("published|\bon:", item.text, re.I):
                        date_matches.append(
                            (datetime_obj, 8)
                        )  # Boost if it has the word published or on
                    else:
                        date_matches.append((datetime_obj, 5))
        candidates = []

        for known_meta_tag in PUBLISH_DATE_META_INFO:
            candidates.extend(parsers.get_metatags(doc, value=known_meta_tag))
        candidates = [(x, "content") for x in candidates]  # property that contains
        # the date
        # is always 'content'
        for known_meta_tag in PUBLISH_DATE_TAGS:
            candidates.extend(
                [
                    (x, known_meta_tag["content"])
                    for x in parsers.get_elements_by_attribs(
                        doc,
                        attribs={known_meta_tag["attribute"]: known_meta_tag["value"]},
                    )
                ]
            )

        for meta_tag, content_attr in candidates:
            date_str = parsers.get_attribute(meta_tag, content_attr)
            datetime_obj = parse_date_str(date_str)
            if datetime_obj:
                score = 6
                if meta_tag.tag.lower() == "meta":
                    score += 1  # Boost meta tags
                days_diff = (datetime.now().date() - datetime_obj.date()).days
                if days_diff < 0:  # articles from the future
                    score -= 2
                elif days_diff > 25 * 365:  # very old articles
                    score -= 1
                date_matches.append((datetime_obj, score))

        date_matches.sort(key=lambda x: x[1], reverse=True)
        self.pubdate = date_matches[0][0] if date_matches else None
        return self.pubdate



# -*- coding: utf-8 -*-
# Much of the code here was forked from https://github.com/codelucas/newspaper
# Copyright (c) Lucas Ou-Yang (codelucas)

from typing import List
import lxml
from newspaper.configuration import Configuration
import newspaper.parsers as parsers
from newspaper.utils.classes import Video

VIDEOS_TAGS = ["iframe", "embed", "object", "video"]
VIDEO_PROVIDERS = ["youtube", "youtu.be", "vimeo", "dailymotion", "kewego", "twitch"]


class VideoExtractor:
    """Extracts a list of video from Article top node"""

    def __init__(self, config: Configuration):
        self.config = config
        self.movies: List[Video] = []

    def parse(self, doc: lxml.html.Element, top_node: lxml.html.Element) -> List[Video]:
        """Extracts video information from the top node

        Args:
            doc (lxml.html.Element): document root
            top_node (lxml.html.Element): Top article node

        Returns:
            List[Video]: List of video objects
        """
        self.movies = []

        if top_node is not None:
            candidates = parsers.get_elements_by_tagslist(top_node, VIDEOS_TAGS)

            for candidate in candidates:
                parser_func = getattr(self, f"parse_{candidate.tag.lower()}")
                if parser_func:
                    video = parser_func(candidate)
                    if video:
                        self.movies.append(video)
        if doc is not None:
            json_ld_scripts = parsers.get_ld_json_object(doc)

            for script_tag in json_ld_scripts:
                if "@graph" in script_tag:
                    script_tag_ = script_tag.get("@graph", [])
                elif isinstance(script_tag, list):
                    script_tag_ = script_tag
                else:
                    script_tag_ = [script_tag]
                for item in script_tag_:
                    if not isinstance(item, dict):
                        continue
                    if item.get("@type") == "VideoObject":
                        m = Video()
                        m.src = item.get("contentUrl")
                        m.embed_code = item.get("embedUrl")
                        m.provider = self._get_provider(m.src)

                        self.movies.append(m)

        return self.movies

    def parse_iframe(self, node: lxml.html.HtmlElement):
        """Parse function for the iframe tag

        Args:
            node (lxml.html.HtmlElement): Input node

        Returns:
            _type_: Video object or None
        """
        return self.parse_video(node)

    def parse_embed(self, node: lxml.html.HtmlElement):
        """Parse function for the embed tag

        Args:
            node (lxml.html.HtmlElement): Input node

        Returns:
            _type_: Video object or None
        """
        parent = node.getparent()
        if parent is not None:
            if parent.tag.lower() == "object":
                return self.parse_object(node)
        return self.parse_video(node)

    def parse_object(self, node: lxml.html.HtmlElement):
        """Parse function for the object tag

        Args:
            node (lxml.html.HtmlElement): Input node

        Returns:
            _type_: Video object or None
        """
        # test if object tag has en embed child
        # in this case we want to remove the embed from
        # the candidate list to avoid parsing it twice
        child_embed_tag = parsers.get_tags(node, "embed")
        if child_embed_tag:
            return None  # Will be parsed as embed

        # get the object source
        # if we don't have a src node don't continue
        src_node = parsers.get_tags(node, tag="param", attribs={"name": "movie"})

        if not src_node:
            return None

        src = parsers.get_attribute(src_node[0], "value")

        # check provider
        provider = self._get_provider(src)
        if not provider:
            return None

        video = self.parse_video(node)
        video.provider = provider
        video.src = src
        return video

    def parse_video(self, node: lxml.html.HtmlElement) -> Video:
        """Parse function for the video tag

        Args:
            node (lxml.html.HtmlElement): Input node

        Returns:
            Video: Video object or None
        """
        video = Video()
        video.embed_code = self._get_embed_code(node)
        video.embed_type = node.tag.lower()
        try:
            video.width = int(node.attrib["width"])
        except (KeyError, ValueError):
            video.width = None
        try:
            video.height = int(node.attrib["height"])
        except (KeyError, ValueError):
            video.height = None
        if node.attrib.get("data-litespeed-src"):
            video.src = node.attrib.get("data-litespeed-src")
        else:
            video.src = node.attrib.get("src")
        video.provider = self._get_provider(video.src)
        return video

    def _get_embed_code(self, node: lxml.html.HtmlElement):
        return "".join(
            [line.strip() for line in parsers.node_to_string(node).splitlines()]
        )

    def _get_provider(self, src: lxml.html.HtmlElement):
        if src:
            for provider in VIDEO_PROVIDERS:
                if provider in src:
                    return provider
        return None
